# coding:utf-8

from urllib.parse import urljoin

__project__ = "xkits"
__version__ = "2.4.0"
__description__ = "Python toolkit."
__urlhome__ = f"https://github.com/bondbox/{__project__}/"
__urlcode__ = __urlhome__
__urldocs__ = __urlhome__
__urlbugs__ = urljoin(__urlhome__, "issues")

# author
__author__ = "Mingzhe Zou"
__author_email__ = "zoumingzhe@outlook.com"

# project name
__prog_name__ = "xkits"
__prog_complete__ = "xargcomplete"
__prog_project__ = "xargproject"
